import sys
import pandas as pd
import typer
from pyaxm.client import Client
from pyaxm.abm_requests import DeviceError

app = typer.Typer()
client = Client()

@app.command()
def devices():
    """List all devices in the organization."""
    df = pd.DataFrame(client.list_devices())
    df.to_csv(sys.stdout, index=False)

@app.command()
def device(device_id: str):
    """Get a device by ID."""
    try:
        df = pd.DataFrame([client.get_device(device_id)])
        df.to_csv(sys.stdout, index=False)
    except DeviceError as e:
        typer.echo(e)

@app.command()
def mdm_servers():
    """List all MDM servers."""
    df = pd.DataFrame(client.list_mdm_servers())
    df.to_csv(sys.stdout, index=False)

@app.command()
def mdm_server(server_id: str):
    """List devices in a specific MDM server."""
    df = pd.DataFrame(client.list_devices_in_mdm_server(server_id))
    df.to_csv(sys.stdout, index=False)

@app.command()
def mdm_server_assigned(device_id: str):
    """Get the server assignment for a device."""
    try:
        df = pd.DataFrame([client.get_device_server_assignment(device_id)])
        df.to_csv(sys.stdout, index=False)
    except DeviceError as e:
        typer.echo(e)

@app.command()
def assign_device(device_id: str, server_id: str):
    """Assign a device to an MDM server."""
    df = pd.DataFrame([client.assign_unassign_device_to_mdm_server(device_id, server_id, 'ASSIGN_DEVICES')])
    df.to_csv(sys.stdout, index=False)

@app.command()
def unassign_device(device_id: str, server_id: str):
    """Unassign a device from an MDM server."""
    df = pd.DataFrame([client.assign_unassign_device_to_mdm_server(device_id, server_id, 'UNASSIGN_DEVICES')])
    df.to_csv(sys.stdout, index=False)

if __name__ == "__main__":
    app()



import datetime as dt
import Cryptodome.PublicKey.ECC as ECC
from authlib.jose import jwt
import uuid
import os
import json
import time
from pyaxm.abm_requests import ABMRequests
from functools import wraps

class AccessToken:
    def __init__(self, value, expires_at):
        self.value = value
        self.expires_at = expires_at

def ensure_valid_token(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self.access_token.expires_at <= dt.datetime.now(dt.timezone.utc):
            self.access_token = self._get_or_refresh_token()
        return method(self, *args, **kwargs)
    return wrapper

class Client:
    def __init__(self, axm_client_id=None, axm_key_id=None, key_path=None, token_path=None):
        # Set configuration from arguments or fall back to environment variables/defaults
        ABM_FOLDER = os.path.join(os.path.expanduser('~'), '.config', 'pyaxm')
        self.axm_client_id = axm_client_id or os.environ.get('AXM_CLIENT_ID')
        self.axm_key_id = axm_key_id or os.environ.get('AXM_KEY_ID')
        self.key_path = key_path or os.path.join(ABM_FOLDER, 'key.pem')
        self.token_path = token_path or os.path.join(ABM_FOLDER, 'token.json')
    	
        self.abm = ABMRequests()
        self.access_token = self._get_or_refresh_token()

    def _get_or_refresh_token(self):
        try:
            with open(self.token_path, 'r') as f:
                cache = json.load(f)
            if cache['expires_at'] > time.time():
                value = cache['access_token']
                expires_at = dt.datetime.fromtimestamp(cache['expires_at'], dt.timezone.utc)
                return AccessToken(value, expires_at)
        except (FileNotFoundError, json.JSONDecodeError, KeyError):
            pass
        # Generate new token
        assertion = self._generate_assertion()
        token_data = {
            "grant_type": "client_credentials",
            "client_id": self.axm_client_id,
            "client_assertion_type": 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            "client_assertion": assertion,
            "scope": 'business.api'
        }
        token = self.abm.get_access_token(token_data)
        expires_in = token.get('expires_in')
        issued_at = dt.datetime.now(dt.timezone.utc)
        expires_at = issued_at + dt.timedelta(seconds=expires_in)
        value = token.get('access_token')
        cache_data = {
            'access_token': value,
            'expires_at': expires_at.timestamp()
        }
        with open(self.token_path, 'w') as f:
            json.dump(cache_data, f)
        return AccessToken(value, expires_at)

    def _generate_assertion(self):
        issued_at = int(dt.datetime.now(dt.timezone.utc).timestamp())
        expires_at = issued_at + 60
        headers = {
            "alg": "ES256",
            "kid": self.axm_key_id
        }
        payload = {
            "sub": self.axm_client_id,
            "aud": 'https://account.apple.com/auth/oauth2/v2/token',
            "iat": issued_at,
            "exp": expires_at,
            "jti": str(uuid.uuid4()),
            "iss": self.axm_client_id
        }
        with open(self.key_path, 'rt') as f:
            private_key = ECC.import_key(f.read())
        assertion = jwt.encode(
            header=headers,
            payload=payload,
            key=private_key.export_key(format='PEM')
        ).decode("utf-8")
        return assertion

    @ensure_valid_token
    def list_devices(self) -> list[dict]:
        response = self.abm.list_devices(self.access_token.value)
        devices = [data.attributes.model_dump() for data in response.data]
        while response.links.next:
            next_page = response.links.next
            response = self.abm.list_devices(self.access_token.value, next=next_page)
            devices.extend([data.attributes.model_dump() for data in response.data])
        return devices

    @ensure_valid_token
    def get_device(self, device_id: str) -> dict:
        response = self.abm.get_device(device_id, self.access_token.value)
        return response.data.attributes.model_dump()
    
    @ensure_valid_token
    def list_mdm_servers(self) -> list[dict]:
        response = self.abm.list_mdm_servers(self.access_token.value)
        exclude_keys = {
            'data': {
                '__all__': {
                    'relationships',
                    'type'
                },
            },
            'included': True,
            'links': True,
            'meta': True,
        }
        data = response.model_dump(exclude=exclude_keys)
        data = [
            {
                'id': server.id,
                'createdDateTime': server.attributes.createdDateTime,
                'serverName': server.attributes.serverName,
                'serverType': server.attributes.serverType,
                'updatedDateTime': server.attributes.updatedDateTime
            } for server in response.data
        ]
        return data

    @ensure_valid_token
    def list_devices_in_mdm_server(self, server_id: str) -> list[str]:
        response = self.abm.list_devices_in_mdm_server(server_id, self.access_token.value)
        include_keys= {
            'data': {
                '__all__': {
                    'id'
                }
            }
        }
        devices = response.model_dump(include=include_keys)
        devices = [device for device in devices['data']]
        while response.links.next:
            next_page = response.links.next
            response = self.abm.list_devices_in_mdm_server(server_id, self.access_token.value, next=next_page)
            devices_dump = response.model_dump(include=include_keys)
            devices_dump = [device for device in devices_dump['data']]
            devices.extend(devices_dump)
        return devices

    @ensure_valid_token
    def get_device_server_assignment(self, device_id: str) -> dict:
        response = self.abm.get_device_server_assignment(device_id, self.access_token.value)
        include_keys = {
            'id'
        }
        return response.data.model_dump(include=include_keys)

    @ensure_valid_token
    def assign_unassign_device_to_mdm_server(
        self, device_id: str, server_id: str, action: str
    ) -> None:
        """
        Assign or unassign a device to/from an MDM server.

        :param device_id: The ID of the device.
        :param server_id: The ID of the MDM server.
        :param action: 'assign' or 'unassign'.
        """
        unassign_response = self.abm.assign_unassign_device_to_mdm_server(
            device_id, server_id, action, self.access_token.value
        )

        # use the ID to check the status until it is complete
        activity_response = self.abm.get_device_activity(unassign_response.data.id, self.access_token.value)
        retry = 0
        while 'COMPLETED' not in activity_response.data.attributes.status and retry < 5:
            time.sleep(2 ** retry)
            retry += 1
            activity_response = self.abm.get_device_activity(unassign_response.data.id, self.access_token.value)

        return activity_response.data.attributes.model_dump()



from pydantic import BaseModel, ConfigDict, AnyHttpUrl, AwareDatetime
from typing import List, Optional, Literal
from enum import Enum

class OrgDeviceActivityType(Enum):
    ASSIGN_DEVICES = "ASSIGN_DEVICES"
    UNASSIGN_DEVICES = "UNASSIGN_DEVICES"

class DocumentLinks(BaseModel):
    self: AnyHttpUrl

class Parameter(BaseModel):
    parameter: str

class JsonPointer(BaseModel):
    pointer: str

class ResourceLinks(BaseModel):
    self: Optional[AnyHttpUrl] = None

class RelationshipLinks(BaseModel):
    include: Optional[AnyHttpUrl] = None
    related: Optional[AnyHttpUrl] = None
    self: Optional[AnyHttpUrl] = None

class PagedDocumentLinks(BaseModel):
    first: Optional[AnyHttpUrl] = None
    next: Optional[AnyHttpUrl] = None
    self: AnyHttpUrl

# OrgDevice
class OrgDevice(BaseModel):
    class Attributes(BaseModel):
        addedToOrgDateTime: Optional[AwareDatetime] = None
        color: Optional[str] = None
        deviceCapacity: Optional[str] = None
        deviceModel: Optional[str] = None
        eid: Optional[str] = None
        imei: Optional[List[str]] = None
        meid: Optional[List[str]] = None
        wifiMacAddress: Optional[str] = None
        bluetoothMacAddress: Optional[str] = None
        orderDateTime: Optional[AwareDatetime] = None
        orderNumber: Optional[str] = None
        partNumber: Optional[str] = None
        productFamily: Optional[str] = None
        productType: Optional[str] = None
        purchaseSourceType: Optional[str] = None
        purchaseSourceId: Optional[str] = None
        serialNumber: Optional[str] = None
        status: Optional[str] = None
        updatedDateTime: Optional[AwareDatetime] = None
    
    class Relationships(BaseModel):
        class AssignedServer(BaseModel):
            links: Optional[RelationshipLinks] = None

        assignedServer: Optional[AssignedServer] = None

    attributes: Optional[Attributes] = None
    id: str
    links: Optional[ResourceLinks] = None
    relationships: Optional[Relationships] = None
    type: Literal['orgDevices']

class OrgDeviceAssignedServerLinkageResponse(BaseModel):
    class Data(BaseModel):
        id: str
        type: Literal['mdmServers']

    data: Data
    links: DocumentLinks

class OrgDeviceActivity(BaseModel):
    class Attributes(BaseModel):
        createdDateTime: Optional[AwareDatetime] = None
        status: Optional[str] = None
        subStatus: Optional[str] = None
        completedDateTime: Optional[AwareDatetime] = None
        downloadUrl: Optional[str] = None

    attributes: Optional[Attributes] = None
    id: str
    links: Optional[ResourceLinks] = None
    type: Literal['orgDeviceActivities']

class OrgDeviceActivityCreateRequest(BaseModel):
    class Data(BaseModel):
        class Attributes(BaseModel):
            activityType: OrgDeviceActivityType
            model_config = ConfigDict(use_enum_values=True)
        
        class Relationships(BaseModel):
            class Devices(BaseModel):
                class Data(BaseModel):
                    id: str
                    type: Literal['orgDevices']
                
                data: List[Data]
            
            class MdmServer(BaseModel):
                class Data(BaseModel):
                    id: str
                    type: Literal['mdmServers']

                data: Data

            devices: Devices
            mdmServer: MdmServer

        attributes: Attributes
        relationships: Relationships
        type: Literal['orgDeviceActivities']
    data: Data

class PagingInformation(BaseModel):
    class Paging(BaseModel):
        limit: int
        nextCursor: Optional[str] = None
        total: Optional[int] = None

    paging: Paging

class MdmServer(BaseModel):
    class Attributes(BaseModel):
        createdDateTime: Optional[AwareDatetime] = None
        serverName: Optional[str] = None
        serverType: Optional[str] = None
        updatedDateTime: Optional[AwareDatetime] = None
    
    class Relationships(BaseModel):
        class Devices(BaseModel):
            class Data(BaseModel):
                id: str
                type: Literal['orgDevices']

            data: Optional[List[Data]] = None
            links: Optional[RelationshipLinks] = None
            meta: Optional[PagingInformation] = None

        devices: Optional[Devices] = None

    attributes: Optional[Attributes] = None
    id: str
    relationships: Optional[Relationships] = None
    type: Literal['mdmServers']

class OrgDeviceActivityResponse(BaseModel):
    data: OrgDeviceActivity
    links: DocumentLinks

class MdmServersResponse(BaseModel):
    data: List[MdmServer]
    included: Optional[List[OrgDevice]] = None
    links: PagedDocumentLinks
    meta: Optional[PagingInformation] = None

class MdmServerResponse(BaseModel):
    data: MdmServer
    included: Optional[List[OrgDevice]] = None
    links: DocumentLinks

class OrgDevicesResponse(BaseModel):
    data: List[OrgDevice]
    links: PagedDocumentLinks
    meta: Optional[PagingInformation] = None

class OrgDeviceResponse(BaseModel):
    data: OrgDevice
    links: DocumentLinks

class ErrorLinks(BaseModel):
    class Associated(BaseModel):
        class Meta(BaseModel):
            source: Optional[str] = None
        
        href: Optional[AnyHttpUrl] = None
        meta: Optional[Meta] = None
    
    about: Optional[AnyHttpUrl] = None
    associated: Optional[AnyHttpUrl|Associated] = None

class ErrorResponse(BaseModel):
    class Errors(BaseModel):
        class Meta(BaseModel):
            # allows non-specified key/value pairs
            model_config = ConfigDict(extra='allow')

        code: str
        detail: str
        id: Optional[str] = None
        source: Optional[JsonPointer|Parameter] = None
        status: str
        title: str
        links: Optional[ErrorLinks] = None
        meta: Optional[Meta] = None
    
    errors: Optional[List[Errors]] = None

class MdmServerDevicesLinkagesResponse(BaseModel):
    class Data(BaseModel):
        id: str
        type: Literal['orgDevices']

    data: List[Data]
    links: PagedDocumentLinks
    meta: Optional[PagingInformation] = None



import requests
from pyaxm.models import (
    OrgDeviceResponse,
    MdmServersResponse,
    MdmServerDevicesLinkagesResponse,
    OrgDevicesResponse,
    OrgDeviceAssignedServerLinkageResponse,
    OrgDeviceActivityCreateRequest,
    OrgDeviceActivityResponse,
)
import time
from functools import wraps

def exponential_backoff(retries=5, backoff_factor=2):
    """
    A decorator for retrying a function with exponential backoff.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for attempt in range(retries):
                try:
                    return func(*args, **kwargs)
                except requests.exceptions.RequestException as e:
                    if attempt < retries - 1:
                        wait_time = backoff_factor ** attempt
                        time.sleep(wait_time)
                    else:
                        raise e
        return wrapper
    return decorator

class DeviceError(Exception):
    pass

class ABMRequests:
    def __init__(self):
        self.session = requests.Session()

    @staticmethod
    def _auth_headers(access_token: str) -> dict:
        """
        :param access_token: The access token for authentication.
        :return: A dictionary containing the authorization headers.
        """
        return {"Authorization": f"Bearer {access_token}"}

    def get_access_token(self, data: dict) -> dict:
        """
        Generate an access token for Apple Business Manager API.
        :param data: A dictionary containing the necessary parameters for the token request.
        :return: A dictionary containing the access token and other related information.
        """
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Host': 'account.apple.com'
        }

        response = self.session.post(
            'https://account.apple.com/auth/oauth2/token',
            headers=headers,
            data=data
        )

        response.raise_for_status()
        return response.json()

    @exponential_backoff(retries=5, backoff_factor=2)
    def list_devices(self, access_token, next=None) -> OrgDevicesResponse:
        """
        List all organization devices.
        :param access_token: The access token for authentication.
        :param next: Optional; the URL for the next page of results.
        :return: An OrgDevicesResponse object containing the list of devices.
        """
        if next:
            url = next
        else:
            url = 'https://api-business.apple.com/v1/orgDevices?limit=1000'

        response = self.session.get(url, headers=self._auth_headers(access_token))

        if response.status_code == 200:
            return OrgDevicesResponse.model_validate(response.json())
        else:
            response.raise_for_status()

    @exponential_backoff(retries=5, backoff_factor=2)
    def get_device(self, device_id, access_token) -> OrgDeviceResponse:
        """
        Retrieve an organization device by its ID.
        
        :param device_id: The ID of the organization device to retrieve.
        :param access_token: The access token for authentication.
        :return: An OrgDeviceResponse object containing the device information.
        """

        url = f'https://api-business.apple.com/v1/orgDevices/{device_id}'
        response = self.session.get(url, headers=self._auth_headers(access_token))
        
        if response.status_code == 200:
            return OrgDeviceResponse.model_validate(response.json())
        elif response.status_code == 404:
            raise DeviceError(response.json()['errors'][0]['title'])
        else:
            response.raise_for_status()

    @exponential_backoff(retries=5, backoff_factor=2)
    def list_mdm_servers(self, access_token) -> MdmServersResponse:
        """
        List all MDM servers.
        
        :param access_token: The access token for authentication.
        :return: An MdmServersResponse object containing the list of MDM servers.
        """
        url = 'https://api-business.apple.com/v1/mdmServers'    
        response = self.session.get(url, headers=self._auth_headers(access_token))
        
        if response.status_code == 200:
            return MdmServersResponse.model_validate(response.json())
        else:
            response.raise_for_status()

    @exponential_backoff(retries=5, backoff_factor=2)
    def list_devices_in_mdm_server(self, server_id: str, access_token, next=None) -> MdmServerDevicesLinkagesResponse:
        """
        List devices in a specific MDM server.
        
        :param server_id: The ID of the MDM server.
        :param access_token: The access token for authentication.
        :param next: Optional; the URL for the next page of results.
        :return: An MdmServerResponse object containing the MDM server information.
        """
        if next:
            url = next
        else:
            url = f'https://api-business.apple.com/v1/mdmServers/{server_id}/relationships/devices?limit=1000'

        response = self.session.get(url, headers=self._auth_headers(access_token))

        # ABM has been returning 500, this is a workaround to retry 2 times
        # before raising an error.
        if response.status_code == 200:
            return MdmServerDevicesLinkagesResponse.model_validate(response.json())
        else:
            response.raise_for_status()

    @exponential_backoff(retries=5, backoff_factor=2)
    def get_device_server_assignment(self, device_id, access_token) -> OrgDeviceAssignedServerLinkageResponse:
        '''Get the server id that a device is assigned to
        '''
        url = f'https://api-business.apple.com/v1/orgDevices/{device_id}/relationships/assignedServer'
        response = self.session.get(url, headers=self._auth_headers(access_token))
        
        if response.status_code == 200:
            return OrgDeviceAssignedServerLinkageResponse.model_validate(response.json())
        elif response.status_code == 404:
            raise DeviceError(response.json()['errors'][0]['title'])
        else:
            response.raise_for_status()

    @exponential_backoff(retries=5, backoff_factor=2)
    def assign_unassign_device_to_mdm_server(
        self, device_id: str, server_id: str, action: str, access_token: str
    ) -> OrgDeviceActivityResponse:
        """
        Assign or unassign a device to/from an MDM server.

        :param device_id: The ID of the device.
        :param server_id: The ID of the MDM server.
        :param action: 'assign' or 'unassign'.
        :param access_token: The access token for authentication.
        """
        url = f'https://api-business.apple.com/v1/orgDeviceActivities'

        request_data = {
            "data": {
            "type": "orgDeviceActivities",
            "attributes": {
                "activityType": action
            },
            "relationships": {
                "devices": {
                "data": [
                    {
                    "id": device_id,
                    "type": "orgDevices"
                    }
                ]
                },
                "mdmServer": {
                "data": {
                    "id": server_id,
                    "type": "mdmServers"
                }
                }
            }
            }
        }

        request_data = OrgDeviceActivityCreateRequest.model_validate(request_data)

        response = self.session.post(
            url,
            headers=self._auth_headers(access_token),
            json=request_data.model_dump()
        )

        if response.status_code == 201:
            return OrgDeviceActivityResponse.model_validate(response.json())
        else:
            response.raise_for_status()

    @exponential_backoff(retries=5, backoff_factor=2)
    def get_device_activity(self, activity_id: str, access_token: str) -> OrgDeviceActivityResponse:
        """
        Get the status of a device activity by its ID.

        :param activity_id: The ID of the device activity.
        :param access_token: The access token for authentication.
        :return: An OrgDeviceActivityResponse object containing the activity information.
        """
        url = f'https://api-business.apple.com/v1/orgDeviceActivities/{activity_id}'
        response = self.session.get(url, headers=self._auth_headers(access_token))

        if response.status_code == 200:
            return OrgDeviceActivityResponse.model_validate(response.json())
        else:
            response.raise_for_status()
